#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "analyze.h"
#include "lib.h"

static int analyze_arp(FILE *fp, char *buffer, int size)
{
  struct arphdr arphdr;
  struct in_addr ipaddr;
  char *p, *smac, *tmac, *sip, *tip;

  if (size < sizeof(arphdr))
    return -1;
  memcpy(&arphdr, buffer, sizeof(arphdr));

  fprintf(fp, "ARP");
  fprintf(fp, "\toperation : %d\n", ntohs(arphdr.ar_op));
  fprintf(fp, "\thardware  : %d\n", ntohs(arphdr.ar_hrd));
  fprintf(fp, "\tprotocol  : 0x%04x\n", ntohs(arphdr.ar_pro));
  fprintf(fp, "\thard size : %d\n", arphdr.ar_hln);
  fprintf(fp, "\tproto size: %d\n", arphdr.ar_pln);

  if ((ntohs(arphdr.ar_hrd) != ARPHRD_ETHER) ||
      (ntohs(arphdr.ar_pro) != ETHERTYPE_IP) ||
      (arphdr.ar_hln != ETHER_ADDR_LEN) ||
      (arphdr.ar_pln != sizeof(in_addr_t)))
    return -1;
  if (size < sizeof(struct arphdr) + arphdr.ar_hln * 2 + arphdr.ar_pln * 2)
    return -1;

  p = buffer;
  p += sizeof(struct arphdr); smac = p;
  p += arphdr.ar_hln; sip  = p;
  p += arphdr.ar_pln; tmac = p;
  p += arphdr.ar_hln; tip  = p;

  fprintf(fp, "\tsender MAC: %s\n", ether_ntoa((struct ether_addr *)smac));
  memcpy(&ipaddr.s_addr, sip, sizeof(in_addr_t));
  fprintf(fp, "\tsender IP : %s\n", inet_ntoa(ipaddr));
  fprintf(fp, "\ttarget MAC: %s\n", ether_ntoa((struct ether_addr *)tmac));
  memcpy(&ipaddr.s_addr, tip, sizeof(in_addr_t));
  fprintf(fp, "\ttarget IP : %s\n", inet_ntoa(ipaddr));

  return 0;
}

static int analyze_icmp(FILE *fp, char *buffer, int size, int total_size)
{
  struct icmp *icmphdr;

  if ((size < total_size) || (total_size < sizeof(*icmphdr)))
    return -1;
  icmphdr = (struct icmp *)buffer;

  fprintf(fp, "ICMP");
  fprintf(fp, "\ttotal size: %d bytes\n", total_size);
  fprintf(fp, "\ttype      : %d\n", icmphdr->icmp_type);
  fprintf(fp, "\tcode      : %d\n", icmphdr->icmp_code);
  fprintf(fp, "\tchecksum  : 0x%04x\n", ntohs(icmphdr->icmp_cksum));

  return 0;
}

static int analyze_igmp(FILE *fp, char *buffer, int size, int total_size)
{
  struct igmp *igmphdr;

  if ((size < total_size) || (total_size < sizeof(*igmphdr)))
    return -1;
  if (total_size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  fprintf(fp, "IGMP");
  fprintf(fp, "\ttype    : %d\n", igmphdr->igmp_type);
  fprintf(fp, "\tcode    : %d\n", igmphdr->igmp_code);
  fprintf(fp, "\tchecksum: 0x%04x\n", ntohs(igmphdr->igmp_cksum));
  fprintf(fp, "\tgroup   : %s\n", inet_ntoa(igmphdr->igmp_group));

  return 0;
}

static int analyze_tcp(FILE *fp, char *buffer, int size, int total_size)
{
  struct tcphdr *tcphdr;

  if ((size < total_size) || (total_size < sizeof(*tcphdr)))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  fprintf(fp, "TCP");
  fprintf(fp, "\ttotal size: %d bytes\n", total_size);
  fprintf(fp, "\tsrc port  : %d\n", ntohs(tcphdr->th_sport));
  fprintf(fp, "\tdst port  : %d\n", ntohs(tcphdr->th_dport));
  fprintf(fp, "\tseq number: %u\n", ntohl(tcphdr->th_seq));
  fprintf(fp, "\tack number: %u\n", ntohl(tcphdr->th_ack));
  fprintf(fp, "\toffset    : %d\n", (tcphdr->th_off << 2));
  fprintf(fp, "\tflags     : 0x%02x ( ", tcphdr->th_flags);
  if (tcphdr->th_flags & TH_FIN ) fprintf(fp, "FIN ");
  if (tcphdr->th_flags & TH_SYN ) fprintf(fp, "SYN ");
  if (tcphdr->th_flags & TH_RST ) fprintf(fp, "RST ");
  if (tcphdr->th_flags & TH_PUSH) fprintf(fp, "PSH ");
  if (tcphdr->th_flags & TH_ACK ) fprintf(fp, "ACK ");
  if (tcphdr->th_flags & TH_URG ) fprintf(fp, "URG ");
  fprintf(fp, ")\n");
  fprintf(fp, "\twindow    : %d\n", ntohs(tcphdr->th_win));
  fprintf(fp, "\tchecksum  : 0x%04x\n", ntohs(tcphdr->th_sum));

  return 0;
}

static int analyze_udp(FILE *fp, char *buffer, int size, int total_size)
{
  struct udphdr *udphdr;

  if ((size < total_size) || (total_size < sizeof(*udphdr)))
    return -1;
  udphdr = (struct udphdr *)buffer;

  fprintf(fp, "UDP");
  fprintf(fp, "\ttotal size: %d bytes\n", total_size);
  fprintf(fp, "\tsrc port  : %d\n", ntohs(udphdr->uh_sport));
  fprintf(fp, "\tdst port  : %d\n", ntohs(udphdr->uh_dport));
  fprintf(fp, "\tlength    : %d\n", ntohs(udphdr->uh_ulen));
  fprintf(fp, "\tchecksum  : 0x%04x\n", ntohs(udphdr->uh_sum));

  return 0;
}

static int analyze_ip(FILE *fp, char *buffer, int size)
{
  int r = -1;
  static char *pktbuf = NULL, *p;
  static int bufsize = 0, s;
  struct ip *iphdr;
  int hdrsize, paysize;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (paysize < 0)
    return -1;
  if (size < hdrsize + paysize)
    return -1;

  fprintf(fp, "IP");
  fprintf(fp, "\theader size: %d bytes\n", hdrsize);
  fprintf(fp, "\ttotal size : %d bytes\n", ntohs(iphdr->ip_len));
  fprintf(fp, "\tID         : 0x%04x\n", ntohs(iphdr->ip_id));
  fprintf(fp, "\tfragment   : 0x%04x\n", ntohs(iphdr->ip_off));
  fprintf(fp, "\tTTL        : %d\n", iphdr->ip_ttl);
  fprintf(fp, "\tprotocol   : %d\n", iphdr->ip_p);
  fprintf(fp, "\tchecksum   : 0x%04x\n", ntohs(iphdr->ip_sum));
  fprintf(fp, "\tsrc IP addr: %s\n", inet_ntoa(iphdr->ip_src));
  fprintf(fp, "\tdst IP addr: %s\n", inet_ntoa(iphdr->ip_dst));

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP: r = analyze_icmp(fp, p, s, paysize); break;
  case IPPROTO_IGMP: r = analyze_igmp(fp, p, s, paysize); break;
  case IPPROTO_TCP:  r = analyze_tcp( fp, p, s, paysize); break;
  case IPPROTO_UDP:  r = analyze_udp( fp, p, s, paysize); break;
  default: break;
  }

  return r;
}

static int analyze_icmp6(FILE *fp, char *buffer, int size, int total_size)
{
  struct icmp6_hdr *icmp6hdr;

  if ((size < total_size) || (total_size < sizeof(*icmp6hdr)))
    return -1;
  icmp6hdr = (struct icmp6_hdr *)buffer;

  fprintf(fp, "ICMPv6");
  fprintf(fp, "\ttotal size: %d bytes\n", total_size);
  fprintf(fp, "\ttype      : %d\n", icmp6hdr->icmp6_type);
  fprintf(fp, "\tcode      : %d\n", icmp6hdr->icmp6_code);
  fprintf(fp, "\tchecksum  : 0x%04x\n", ntohs(icmp6hdr->icmp6_cksum));

  return 0;
}

static int analyze_ip6(FILE *fp, char *buffer, int size)
{
  int r = -1;
  static char *pktbuf = NULL, *p;
  static int bufsize = 0, s;
  struct ip6_hdr *ip6hdr;
  int hdrsize, paysize;
  int nexthdr, exthdrsize;
  struct ip6_ext *ip6exthdr;
  struct ip6_rthdr *ip6rthdr;
  struct ip6_frag *ip6fraghdr;
  char addrs[INET6_ADDRSTRLEN];

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*ip6hdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  ip6hdr = (struct ip6_hdr *)pktbuf;

  hdrsize = sizeof(struct ip6_hdr);
  paysize = ntohs(ip6hdr->ip6_plen);

  if (size < hdrsize + paysize)
    return -1;

  fprintf(fp, "IPv6");
  fprintf(fp, "\tpayload size : %d bytes\n", paysize);
  fprintf(fp, "\thop limit    : %d\n", ip6hdr->ip6_hlim);
  if (inet_ntop(AF_INET6, &ip6hdr->ip6_src, addrs, sizeof(addrs)) != NULL)
    fprintf(fp, "\tsrc IPv6 addr: %s\n", addrs);
  if (inet_ntop(AF_INET6, &ip6hdr->ip6_dst, addrs, sizeof(addrs)) != NULL)
    fprintf(fp, "\tdst IPv6 addr: %s\n", addrs);

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  nexthdr = ip6hdr->ip6_nxt;

  while (1) {
    fprintf(fp, "\tnext header  : %d\n", nexthdr);
    exthdrsize = 0;
    switch (nexthdr) {
    case IPPROTO_HOPOPTS:
    case IPPROTO_DSTOPTS:
      if (paysize < sizeof(*ip6exthdr))
	return -1;
      ip6exthdr = (struct ip6_ext *)p;
      nexthdr = ip6exthdr->ip6e_nxt;
      exthdrsize = (ip6exthdr->ip6e_len + 1) << 3;
      break;
    case IPPROTO_ROUTING:
      if (paysize < sizeof(*ip6rthdr))
	return -1;
      ip6rthdr = (struct ip6_rthdr *)p;
      nexthdr = ip6rthdr->ip6r_nxt;
      exthdrsize = (ip6rthdr->ip6r_len + 1) << 3;
      break;
    case IPPROTO_FRAGMENT:
      if (paysize < sizeof(*ip6fraghdr))
	return -1;
      ip6fraghdr = (struct ip6_frag *)p;
      nexthdr = ip6fraghdr->ip6f_nxt;
      exthdrsize = sizeof(*ip6fraghdr);
      break;
    case IPPROTO_NONE:
    default:
      break;
    }
    if (exthdrsize == 0)
      break;
    if (paysize < exthdrsize)
      return -1;
    fprintf(fp, "\texthdr size  : %d bytes\n", exthdrsize);
    p       += exthdrsize;
    s       -= exthdrsize;
    paysize -= exthdrsize;
  }

  switch (nexthdr) {
  case IPPROTO_ICMPV6: r = analyze_icmp6(fp, p, s, paysize); break;
  case IPPROTO_TCP:    r = analyze_tcp(  fp, p, s, paysize); break;
  case IPPROTO_UDP:    r = analyze_udp(  fp, p, s, paysize); break;
  default: break;
  }

  return r;
}

int pkt_analyze_ethernet(FILE *fp, char *buffer, int size, struct timeval *tm)
{
  int r = -1;
  struct ether_header ehdr;
  struct {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;
  static int count = 0;

  count++;
  fprintf(fp, "-- %d --\n", count);
  fprintf(fp, "received: %d bytes\n", size);
  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);
  fprintf(fp, "%s ->", ether_ntoa((struct ether_addr *)ehdr.ether_shost));
  fprintf(fp, " %s  ", ether_ntoa((struct ether_addr *)ehdr.ether_dhost));

  type = ntohs(ehdr.ether_type);
  buffer += ETHER_HDR_LEN;
  size   -= ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86DD
#endif

  while (type == ETHERTYPE_VLAN) {
    if (size < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, buffer, sizeof(vlantag));
    fprintf(fp, "(VLAN tag: 0x%04x) ", ntohs(vlantag.tag));
    type = ntohs(vlantag.proto);
    buffer += sizeof(vlantag);
    size   -= sizeof(vlantag);
  }

  fprintf(fp, "(type: 0x%04x)\n", type);

  switch (type) {
  case ETHERTYPE_ARP:  r = analyze_arp(fp, buffer, size); break;
  case ETHERTYPE_IP:   r = analyze_ip( fp, buffer, size); break;
  case ETHERTYPE_IPV6: r = analyze_ip6(fp, buffer, size); break;
  default: break;
  }

  fprintf(fp, "==\n");
  fflush(fp);

  return r;
}
