#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "correct.h"
#include "lib.h"

struct pseudo_header {
  in_addr_t saddr;
  in_addr_t daddr;
  pkt_uint8 zero;
  pkt_uint8 protocol;
  pkt_uint16 len;
};

struct pseudo6_header {
  struct in6_addr saddr;
  struct in6_addr daddr;
  pkt_uint32 len;
  pkt_uint8 zero[3];
  pkt_uint8 nexthdr;
};

static int correct_icmp(char *buffer, int size, int total_size)
{
  struct icmp *icmphdr;

  if ((size < total_size) || (total_size < sizeof(*icmphdr)))
    return -1;
  icmphdr = (struct icmp *)buffer;

  /* This is compatible with FreeBSD */
  icmphdr->icmp_cksum = 0;
  icmphdr->icmp_cksum = htons(~ip_checksum(icmphdr, total_size));

  return 0;
}

static int correct_igmp(char *buffer, int size, int total_size)
{
  struct igmp *igmphdr;

  if ((size < total_size) || (total_size < sizeof(*igmphdr)))
    return -1;
  if (total_size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  /* This is compatible with FreeBSD */
  igmphdr->igmp_cksum = 0;
  igmphdr->igmp_cksum = htons(~ip_checksum(igmphdr, sizeof(struct igmp)));

  return 0;
}

static int correct_tcp(char *buffer, int size, int total_size, int pchksum)
{
  struct tcphdr *tcphdr;

  if ((size < total_size) || (total_size < sizeof(*tcphdr)))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  /* This is compatible with FreeBSD */
  tcphdr->th_sum = htons(pchksum);
  tcphdr->th_sum = htons(~ip_checksum(tcphdr, total_size));

  return 0;
}

static int correct_udp(char *buffer, int size, int total_size, int pchksum)
{
  struct udphdr *udphdr;

  if ((size < total_size) || (total_size < sizeof(*udphdr)))
    return -1;
  udphdr = (struct udphdr *)buffer;

  /* This is compatible with FreeBSD */
  udphdr->uh_sum = htons(pchksum);
  udphdr->uh_sum = htons(~ip_checksum(udphdr, total_size));

  return 0;
}

static int correct_ip(char *buffer, int size)
{
  int r = -1;
  static char *pktbuf = NULL, *p;
  static int bufsize = 0, s;
  struct ip *iphdr;
  int hdrsize, paysize;
  struct pseudo_header phdr;
  int pchksum;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (paysize < 0)
    return -1;
  if (size < hdrsize + paysize)
    return -1;

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  memset(&phdr, 0, sizeof(phdr));
  phdr.saddr = iphdr->ip_src.s_addr;
  phdr.daddr = iphdr->ip_dst.s_addr;
  phdr.protocol = iphdr->ip_p;
  phdr.len = htons(paysize);
  pchksum = ip_checksum(&phdr, sizeof(phdr));

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP: r = correct_icmp(p, s, paysize); break;
  case IPPROTO_IGMP: r = correct_igmp(p, s, paysize); break;
  case IPPROTO_TCP:  r = correct_tcp( p, s, paysize, pchksum); break;
  case IPPROTO_UDP:  r = correct_udp( p, s, paysize, pchksum); break;
  default: break;
  }

  /* This is compatible with FreeBSD */
  iphdr->ip_sum = 0;
  iphdr->ip_sum = htons(~ip_checksum(iphdr, hdrsize));

  memcpy(buffer, pktbuf, size);

  return r;
}

static int correct_icmp6(char *buffer, int size, int total_size, int pchksum)
{
  struct icmp6_hdr *icmp6hdr;

  if ((size < total_size) || (total_size < sizeof(*icmp6hdr)))
    return -1;
  icmp6hdr = (struct icmp6_hdr *)buffer;

  /* This is compatible with FreeBSD */
  icmp6hdr->icmp6_cksum = htons(pchksum);
  icmp6hdr->icmp6_cksum = htons(~ip_checksum(icmp6hdr, total_size));

  return 0;
}

static int correct_ip6(char *buffer, int size)
{
  int r = -1;
  static char *pktbuf = NULL, *p;
  static int bufsize = 0, s;
  struct ip6_hdr *ip6hdr;
  int hdrsize, paysize;
  int nexthdr, exthdrsize;
  struct ip6_ext *ip6exthdr;
  struct ip6_rthdr *ip6rthdr;
  struct ip6_frag *ip6fraghdr;
  struct pseudo6_header phdr;
  int pchksum;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*ip6hdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  ip6hdr = (struct ip6_hdr *)pktbuf;

  hdrsize = sizeof(struct ip6_hdr);
  paysize = ntohs(ip6hdr->ip6_plen);

  if (size < hdrsize + paysize)
    return -1;

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  nexthdr = ip6hdr->ip6_nxt;

  while (1) {
    exthdrsize = 0;
    switch (nexthdr) {
    case IPPROTO_HOPOPTS:
    case IPPROTO_DSTOPTS:
      if (paysize < sizeof(*ip6exthdr))
	return -1;
      ip6exthdr = (struct ip6_ext *)p;
      nexthdr = ip6exthdr->ip6e_nxt;
      exthdrsize = (ip6exthdr->ip6e_len + 1) << 3;
      break;
    case IPPROTO_ROUTING:
      if (paysize < sizeof(*ip6rthdr))
	return -1;
      ip6rthdr = (struct ip6_rthdr *)p;
      nexthdr = ip6rthdr->ip6r_nxt;
      exthdrsize = (ip6rthdr->ip6r_len + 1) << 3;
      break;
    case IPPROTO_FRAGMENT:
      if (paysize < sizeof(*ip6fraghdr))
	return -1;
      ip6fraghdr = (struct ip6_frag *)p;
      nexthdr = ip6fraghdr->ip6f_nxt;
      exthdrsize = sizeof(*ip6fraghdr);
      break;
    case IPPROTO_NONE:
    default:
      break;
    }
    if (exthdrsize == 0)
      break;
    if (paysize < exthdrsize)
      return -1;
    p       += exthdrsize;
    s       -= exthdrsize;
    paysize -= exthdrsize;
  }

  memset(&phdr, 0, sizeof(phdr));
  memcpy(&phdr.saddr, &ip6hdr->ip6_src, sizeof(struct in6_addr));
  memcpy(&phdr.daddr, &ip6hdr->ip6_dst, sizeof(struct in6_addr));
  phdr.len = htonl(paysize);
  phdr.nexthdr = nexthdr;
  pchksum = ip_checksum(&phdr, sizeof(phdr));

  switch (nexthdr) {
  case IPPROTO_ICMPV6: r = correct_icmp6(p, s, paysize, pchksum); break;
  case IPPROTO_TCP:    r = correct_tcp(  p, s, paysize, pchksum); break;
  case IPPROTO_UDP:    r = correct_udp(  p, s, paysize, pchksum); break;
  default: break;
  }

  memcpy(buffer, pktbuf, size);

  return r;
}

int pkt_correct_ethernet(char *buffer, int size)
{
  int r = -1;
  struct ether_header ehdr;
  struct {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;

  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);

  type = ntohs(ehdr.ether_type);
  buffer += ETHER_HDR_LEN;
  size   -= ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86DD
#endif

  while (type == ETHERTYPE_VLAN) {
    if (size < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, buffer, sizeof(vlantag));
    type = ntohs(vlantag.proto);
    buffer += sizeof(vlantag);
    size   -= sizeof(vlantag);
  }

  switch (type) {
  case ETHERTYPE_IP:   r = correct_ip( buffer, size); break;
  case ETHERTYPE_IPV6: r = correct_ip6(buffer, size); break;
  default: break;
  }

  return r;
}
