#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>

#include "pingrep.h"
#include "lib.h"

static int make_srcmacaddr(char *macaddr, struct in_addr *ipaddr)
{
  macaddr[0] = 0x00;
  macaddr[1] = 0x11;
  memcpy(&macaddr[2], &ipaddr->s_addr, sizeof(in_addr_t));
  return 0;
}

static int memswap(void *p0, void *p1, int size)
{
  char *c0 = p0, *c1 = p1;
  char c;
  while (size > 0) {
    c = *c0; *c0 = *c1; *c1 = c;
    c0++; c1++; size--;
  }
  return 0;
}

static int pingrep_arp(char *buffer, int size, char *macaddr)
{
  struct arphdr arphdr;
  struct in_addr ipaddr;
  char *p, *smac, *tmac, *sip, *tip;

  if (size < sizeof(arphdr))
    return -1;
  memcpy(&arphdr, buffer, sizeof(arphdr));

  if ((ntohs(arphdr.ar_hrd) != ARPHRD_ETHER) ||
      (ntohs(arphdr.ar_pro) != ETHERTYPE_IP) ||
      (arphdr.ar_hln != ETHER_ADDR_LEN) ||
      (arphdr.ar_pln != sizeof(in_addr_t)))
    return -1;
  if (size < sizeof(struct arphdr) + arphdr.ar_hln * 2 + arphdr.ar_pln * 2)
    return -1;

  if (ntohs(arphdr.ar_op) != ARPOP_REQUEST)
    return -1;

  p = buffer;
  p += sizeof(struct arphdr); smac = p;
  p += arphdr.ar_hln; sip  = p;
  p += arphdr.ar_pln; tmac = p;
  p += arphdr.ar_hln; tip  = p;

  memswap(sip, tip, sizeof(in_addr_t));
  memcpy(tmac, smac, ETHER_ADDR_LEN);
  memcpy(&ipaddr.s_addr, sip, sizeof(in_addr_t));
  make_srcmacaddr(smac, &ipaddr);
  memcpy(macaddr, smac, ETHER_ADDR_LEN);

  arphdr.ar_op = htons(ARPOP_REPLY);

  memcpy(buffer, &arphdr, sizeof(arphdr));

  return 0;
}

static int pingrep_icmp(char *buffer, int size, int total_size)
{
  struct icmp *icmphdr;

  if ((size < total_size) || (total_size < sizeof(*icmphdr)))
    return -1;
  icmphdr = (struct icmp *)buffer;

  if (icmphdr->icmp_type != ICMP_ECHO)
    return -1;

  icmphdr->icmp_type = ICMP_ECHOREPLY;

  /* This is compatible with FreeBSD */
  icmphdr->icmp_cksum = 0;
  icmphdr->icmp_cksum = htons(~ip_checksum(icmphdr, total_size));

  return 0;
}

static int pingrep_ip(char *buffer, int size, char *macaddr)
{
  int r = -1;
  static char *pktbuf = NULL, *p;
  static int bufsize = 0, s;
  struct ip *iphdr;
  int hdrsize, paysize;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (paysize < 0)
    return -1;
  if (size < hdrsize + paysize)
    return -1;

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP: r = pingrep_icmp(p, s, paysize); break;
  default: break;
  }

  if (r < 0)
    return -1;

  iphdr->ip_ttl = 255;
  memswap(&iphdr->ip_src, &iphdr->ip_dst, sizeof(in_addr_t));

  /* This is compatible with FreeBSD */
  iphdr->ip_sum = 0;
  iphdr->ip_sum = htons(~ip_checksum(iphdr, hdrsize));

  make_srcmacaddr(macaddr, &iphdr->ip_src);
  memcpy(buffer, pktbuf, size);

  return r;
}

int pkt_pingrep_ethernet(char *buffer, int size)
{
  int r = -1;
  struct ether_header ehdr, *ehdrp;
  struct {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;
  char smac[ETHER_ADDR_LEN];

  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);
  ehdrp = (struct ether_header *)buffer;

  type = ntohs(ehdr.ether_type);
  buffer += ETHER_HDR_LEN;
  size   -= ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif

  while (type == ETHERTYPE_VLAN) {
    if (size < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, buffer, sizeof(vlantag));
    type = ntohs(vlantag.proto);
    buffer += sizeof(vlantag);
    size   -= sizeof(vlantag);
  }

  switch (type) {
  case ETHERTYPE_ARP: r = pingrep_arp(buffer, size, smac); break;
  case ETHERTYPE_IP:  r = pingrep_ip( buffer, size, smac); break;
  default: break;
  }

  if (r < 0)
    return -1;

  memcpy(ehdrp->ether_dhost, ehdrp->ether_shost, ETHER_ADDR_LEN);
  memcpy(ehdrp->ether_shost, smac, ETHER_ADDR_LEN);

  return r;
}
