#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>

#include "text.h"
#include "lib.h"

#define COLUMN 16

static int isterminator(int c)
{
  return ((c == EOF) || (c == '\r') || (c == '\n')) ? 1 : 0;
}

static int getkey(FILE *fp, char *buffer, int size)
{
  int c, i = 0;

  while ((c = fgetc(fp)) != EOF) {
    if ((i >= size - 1) || (c == EOF) || (c == ':') || isspace(c))
      break;
    buffer[i++] = c;
  }
  buffer[i] = '\0';

  return c;
}

static int skip_line(FILE *fp)
{
  int c;
  while ((c = fgetc(fp)) != EOF) {
    if (isterminator(c))
      break;
  }
  return c;
}

static int read_num(FILE *fp, int *nump)
{
  int c, n, num = -1;
  while ((c = fgetc(fp)) != EOF) {
    if (num < 0) {
      if (isspace(c) && !isterminator(c))
	continue;
      if (!isdigit(c))
	break;
      num = 0;
    } else if (!isdigit(c)) {
      break;
    }
    n = c - '0';
    num = num * 10 + n;
  }
  if (nump) *nump = num;
  return c;
}

static int read_hex(FILE *fp, int *nump)
{
  int c, n, num = -1;
  while ((c = fgetc(fp)) != EOF) {
    if (num < 0) {
      if (isspace(c) && !isterminator(c))
	continue;
      if (!isxdigit(c))
	break;
      num = 0;
    } else if (!isxdigit(c)) {
      break;
    }
    if (isdigit(c))
      n = c - '0';
    else if ((c >= 'a') && (c <= 'f'))
      n = c - 'a' + 10;
    else
      n = c - 'A' + 10;
    num = num * 16 + n;
  }
  if (nump) *nump = num;
  return c;
}

int pkt_text_read(FILE *fp, char *buffer, int size,
		  int *capsizep, int *origsizep, struct timeval *tp)
{
  char inbuf[32];
  int i = 0, c, n, capsize, origsize;
  struct timeval t;

  capsize = origsize = -1;
  t.tv_sec = t.tv_usec = 0;

  while (1) {
    c = getkey(fp, inbuf, sizeof(inbuf));
    if (c == EOF)
      return -1;
    if (!strcmp(inbuf, "--")) {
      if (!isterminator(c))
	c = skip_line(fp);
      if (c == EOF)
	return -1;
      break;
    }
  }

  while (1) {
    c = getkey(fp, inbuf, sizeof(inbuf));

    if (!strcmp(inbuf, "==")) {
      if (!isterminator(c))
	c = skip_line(fp);
      break;
    }

    if (c == EOF)
      return -1;

    if (!strcmp(inbuf, "TIME")) {
      c = read_num(fp, &n);
      if (n >= 0) t.tv_sec = n;
      if (c == '.') {
	if (n < 0) t.tv_sec = 0;
	c = read_num(fp, &n);
	if (n >= 0) t.tv_usec = n;
	else t.tv_usec = 0;
      } else {
	if (n >= 0)
	  t.tv_usec = 0;
      }
      if (!isterminator(c))
	c = skip_line(fp);
      if (c == EOF)
	return -1;
      continue;
    }

    if (!strcmp(inbuf, "SIZE")) {
      c = read_num(fp, &n);
      if (n >= 0) capsize = n;
      if (c == '/') {
	if (n < 0) capsize = 0;
	c = read_num(fp, &n);
	if (n >= 0) origsize = n;
	else origsize = 0;
      } else {
	if (n >= 0)
	  origsize = -1;
      }
      if (!isterminator(c))
	c = skip_line(fp);
      if (c == EOF)
	return -1;
      continue;
    }

    if (isspace(c)) {
      if (i < size)
	buffer[i] = strtoul(inbuf, NULL, 16);
      else
	error_exit("Out of buffer.\n");
      i++;
    }
    while (1) {
      c = read_hex(fp, &n);
      if (n >= 0) {
	if (i < size)
	  buffer[i] = n;
	else
	  error_exit("Out of buffer.\n");
	i++;
      }
      if ((n < 0) || !isspace(c) || isterminator(c)) {
	if (!isterminator(c))
	  c = skip_line(fp);
	if (c == EOF)
	  return -1;
	break;
      }
    }
  }

  if ((capsize < 0) || (capsize > i))
    capsize = i;
  if (origsize < 0)
    origsize = capsize;

  if (capsize >= 0) {
    if (capsizep ) *capsizep  = capsize;
    if (origsizep) *origsizep = origsize;
    if (tp) {
      tp->tv_sec  = t.tv_sec;
      tp->tv_usec = t.tv_usec;
    }
  }

  return capsize;
}

int pkt_text_write(FILE *fp, char *buffer,
		   int capsize, int origsize, struct timeval *tp)
{
  char text[COLUMN * 2];
  char *textp = NULL;
  unsigned char c;
  struct timeval zero;
  int i, size;
  static int count = 0;

  if (origsize < 0)
    origsize = capsize;

  zero.tv_sec = zero.tv_usec = 0;
  if (tp == NULL)
    tp = &zero;

  count++;
  fprintf(fp, "-- %d --\n", count);
  fprintf(fp, "TIME: %d.%d %s",
	  (int)tp->tv_sec, (int)tp->tv_usec, ctime(&tp->tv_sec));
  fprintf(fp, "SIZE: %d/%d\n", capsize, origsize);

  size = ((capsize + COLUMN - 1) / COLUMN) * COLUMN;
  for (i = 0; i < size; i++) {
    if ((i % COLUMN) == 0) {
      textp = text;
      fprintf(fp, "%06X:", i);
    }

    if ((i % 4) == 0)
      fprintf(fp, " ");

    if (i < capsize) {
      c = buffer[i];
      fprintf(fp, "%02X ", c);
    } else {
      c = ' ';
      fprintf(fp, "   ");
    }

    if ((i % 8) == 0) *(textp++) = ' ';
    *(textp++) = isprint(c) ? c : '.';
    *textp = '\0';

    if ((i % COLUMN) == (COLUMN - 1))
      fprintf(fp, ":%s\n", text);
  }

  fprintf(fp, "==\n");
  fflush(fp);

  return 0;
}
